from __future__ import annotations

from dataclasses import dataclass

import numpy as np

from ftag import Cuts


@dataclass
class TrackSelector:
    """
    Apply track selections to a set of tracks stored in a structured numpy array.

    The array is assumed to have shape (n_jets, n_tracks, n_features).
    Applying cuts will NaN out the tracks that do not pass the cuts,
    but leave the shape of the array unchanged.

    Parameters
    ----------
    cuts : Cuts
        The cuts to apply to the tracks
    valid_str : str
        The name of the field in the tracks that indicates whether the track is
    """

    cuts: Cuts
    valid_str: str = "valid"

    def __call__(self, tracks: np.ndarray) -> np.ndarray:
        # get a bool array for all tracks passing before any cuts
        rm_idx = np.zeros_like(tracks[self.valid_str], dtype=bool)

        # apply the cuts
        for cut in self.cuts.cuts:
            # remove valid track indices that do not pass the selection
            rm_idx[tracks[self.valid_str] & ~cut(tracks)] = True

        # set the values of the tracks that do not pass the cuts to
        for var in tracks.dtype.names:
            if issubclass(tracks[var].dtype.type, np.floating):
                tracks[var][rm_idx] = np.nan
            elif issubclass(tracks[var].dtype.type, np.integer):
                tracks[var][rm_idx] = -1
            elif issubclass(tracks[var].dtype.type, np.bool_):
                tracks[var][rm_idx] = False
            else:
                raise TypeError(f"Unknown dtype {tracks[var].dtype}")

        # specifically set the valid flag to false (even though it's already false by now)
        tracks[rm_idx][self.valid_str] = False

        return tracks



from __future__ import annotations

from dataclasses import dataclass

from ftag.cuts import Cuts


@dataclass(frozen=True)
class Region:
    name: str
    cuts: Cuts

    def __str__(self) -> str:
        return self.name

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return self.name == other.name



from __future__ import annotations

from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory

import pytest

from ftag.cli_utils import valid_path


def test_valid_path_existing_file():
    # Test when the input path is an existing file
    # get a temp directory
    with TemporaryDirectory() as tmpdir, NamedTemporaryFile(dir=tmpdir) as f:
        input_path = f.name
        expected_output = Path(f.name)
        result = valid_path(input_path)
        assert result == expected_output


def test_valid_path_non_existing_file():
    # Test when the input path is a non-existing file
    input_path = "non_existing_file.txt"
    with pytest.raises(FileNotFoundError) as e:
        valid_path(input_path)
    assert str(e.value) == input_path


def test_valid_path_directory():
    # Test when the input path is a directory
    input_path = "directory/"
    with pytest.raises(FileNotFoundError) as e:
        valid_path(input_path)
    assert str(e.value) == input_path



from __future__ import annotations

import argparse
from pathlib import Path


def valid_path(string):
    if (path := Path(string)).is_file():
        return path
    raise FileNotFoundError(string)


class HelpFormatter(argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter): ...



from __future__ import annotations

import glob
from dataclasses import dataclass
from pathlib import Path

from ftag.flavour import remove_suffix
from ftag.vds import create_virtual_file


@dataclass(frozen=True)
class Sample:
    pattern: Path | str | tuple[Path | str, ...]
    ntuple_dir: Path | str | None = None
    name: str | None = None

    def __post_init__(self) -> None:
        if not self.pattern:
            raise ValueError("Sample pattern cannot be empty")
        if "*" in str(self.pattern) and not self.files:
            raise FileNotFoundError(f"No files matched pattern {self.pattern}")
        if missing := [file for file in self.files if not Path(file).is_file()]:
            raise FileNotFoundError(f"The following files do not exist: {missing}")

    @property
    def path(self) -> tuple[Path, ...]:
        pattern_tuple = self.pattern if isinstance(self.pattern, (list, tuple)) else (self.pattern,)
        if self.ntuple_dir is not None:
            return tuple(Path(self.ntuple_dir, p) for p in pattern_tuple)
        return tuple(Path(p) for p in pattern_tuple)

    @property
    def files(self) -> list[str]:
        files = []
        for p in self.path:
            files += glob.glob(str(p)) if "*" in str(p) else [str(p)]
        return files

    @property
    def num_files(self) -> int:
        return len(self.files)

    @property
    def dsid(self) -> list[str]:
        return list({Path(fname).parent.name for fname in self.files})

    @property
    def sample_id(self) -> list[str]:
        return list({dsid.split(".")[2] for dsid in self.dsid})

    @property
    def tags(self) -> list[str]:
        return list({dsid.split(".")[3] for dsid in self.dsid})

    @property
    def ptag(self) -> list[str]:
        return list({tag for tags in self.tags for tag in tags.split("_") if "p" in tag})

    @property
    def rtag(self) -> list[str]:
        return list({tag for tags in self.tags for tag in tags.split("_") if "r" in tag})

    @property
    def dumper_tag(self) -> list[str]:
        hashes = [remove_suffix(dsid.split(".")[7], "_output") for dsid in self.dsid]
        return list(set(hashes))

    def virtual_file(self, **kwargs) -> list[Path | str]:
        return [create_virtual_file(p, **kwargs) if "*" in str(p) else p for p in self.path]

    def __str__(self):
        return self.name

    def __lt__(self, other):
        return self.name < other.name

    def __eq__(self, other):
        return self.name == other.name



from __future__ import annotations

import functools
import operator
from ast import literal_eval
from collections.abc import Iterator
from dataclasses import dataclass
from itertools import starmap
from typing import NamedTuple

import numpy as np

OPERATORS = {
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    "<=": operator.le,
    ">": operator.gt,
    "<": operator.lt,
    "in": np.isin,
    "notin": lambda x, y: ~np.isin(x, y),
}

for i in range(2, 101):
    OPERATORS[f"%{i}=="] = functools.partial(lambda x, y, i: (x % i) == y, i=i)
    OPERATORS[f"%{i}!="] = functools.partial(lambda x, y, i: (x % i) != y, i=i)
    OPERATORS[f"%{i}<="] = functools.partial(lambda x, y, i: (x % i) <= y, i=i)
    OPERATORS[f"%{i}>="] = functools.partial(lambda x, y, i: (x % i) >= y, i=i)


class CutsResult(NamedTuple):
    idx: np.ndarray
    values: np.ndarray


@dataclass(frozen=True)
class Cut:
    variable: str
    operator: str
    _value: str | int | float

    @property
    def value(self) -> int | float:
        if isinstance(self._value, str):
            return literal_eval(self._value)
        return self._value

    def __call__(self, array):
        return OPERATORS[self.operator](array[self.variable], self.value)

    def __str__(self) -> str:
        return f"{self.variable} {self.operator} {self.value}"


@dataclass(frozen=True)
class Cuts:
    cuts: tuple[Cut, ...]

    @classmethod
    def from_list(cls, cuts: list) -> Cuts:
        if cuts and isinstance(cuts[0], str):
            cuts = [cut.split(" ") for cut in cuts]
        if cuts and isinstance(cuts[0], list):
            cuts = list(map(tuple, cuts))
        return cls(tuple(starmap(Cut, dict.fromkeys(cuts))))

    @classmethod
    def empty(cls) -> Cuts:
        return cls(())

    def __post_init__(self):
        assert isinstance(self.cuts, tuple)
        assert all(isinstance(c, Cut) for c in self.cuts)

    @property
    def variables(self) -> list[str]:
        return list(dict.fromkeys(c.variable for c in self))

    def ignore(self, variables: list[str]):
        return Cuts(tuple(c for c in self if c.variable not in variables))

    def __call__(self, array: np.ndarray) -> CutsResult:
        if array.ndim == 2:
            raise ValueError("This interface only supports jet selections")
        keep = np.arange(len(array))
        for cut in self.cuts:
            idx = cut(array)
            array, keep = array[idx], keep[idx]
        return CutsResult(keep, array)

    def __add__(self, other: Cuts):
        return Cuts(tuple(dict.fromkeys(self.cuts + other.cuts)))

    def __len__(self) -> int:
        return len(self.cuts)

    def __iter__(self) -> Iterator:
        yield from self.cuts

    def __getitem__(self, variable):
        return Cuts(tuple(cut for cut in self.cuts if cut.variable == variable))

    def __repr__(self) -> str:
        return str([f"{c}" for c in self.cuts])



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import Path

import yaml

from ftag.cuts import Cuts


def remove_suffix(string: str, suffix: str) -> str:
    if string.endswith(suffix):
        return string[: -len(suffix)]
    return string


@dataclass(frozen=True)
class Flavour:
    name: str
    label: str
    cuts: Cuts
    colour: str
    category: str

    @property
    def px(self) -> str:
        return f"p{remove_suffix(self.name, 'jets')}"

    @property
    def eff_str(self) -> str:
        return self.label.replace("jets", "jet") + " efficiency"

    @property
    def rej_str(self) -> str:
        return self.label.replace("jets", "jet") + " rejection"

    @property
    def frac_str(self) -> str:
        return "f" + remove_suffix(self.name, "jets")

    def __str__(self) -> str:
        return self.name

    def __lt__(self, other) -> bool:
        return self.name < other.name


@dataclass
class FlavourContainer:
    flavours: dict[str, Flavour]

    def __iter__(self) -> Iterator:
        yield from self.flavours.values()

    def __getitem__(self, key) -> Flavour:
        if isinstance(key, Flavour):
            key = key.name
        try:
            return self.flavours[key]
        except KeyError as e:
            raise KeyError(f"Flavour '{key}' not found") from e

    def __getattr__(self, name) -> Flavour:
        return self[name]

    def __contains__(self, flavour: str | Flavour) -> bool:
        if isinstance(flavour, Flavour):
            flavour = flavour.name
        return flavour in self.flavours

    def __eq__(self, other) -> bool:
        if isinstance(other, FlavourContainer):
            return self.flavours == other.flavours
        if isinstance(other, list) and all(isinstance(f, str) for f in other):
            return {f.name for f in self} == set(other)
        return False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({', '.join([f.name for f in self])})"

    @property
    def categories(self) -> list[str]:
        return list(dict.fromkeys(f.category for f in self))

    def by_category(self, category: str) -> FlavourContainer:
        f = FlavourContainer({k: v for k, v in self.flavours.items() if v.category == category})
        if not f.flavours:
            raise KeyError(f"No flavours with category '{category}' found")
        return f

    def from_cuts(self, cuts: list | Cuts) -> Flavour:
        if isinstance(cuts, list):
            cuts = Cuts.from_list(cuts)
        for flavour in self:
            if flavour.cuts == cuts:
                return flavour
        raise KeyError(f"Flavour with {cuts} not found")

    @classmethod
    def from_yaml(cls, yaml_path: Path | None = None) -> FlavourContainer:
        if yaml_path is None:
            yaml_path = Path(__file__).parent / "flavours.yaml"

        with open(yaml_path) as f:
            flavours_yaml = yaml.safe_load(f)

        flavours_dict = {
            f["name"]: Flavour(cuts=Cuts.from_list(f.pop("cuts")), **f) for f in flavours_yaml
        }
        assert len(flavours_dict) == len(flavours_yaml), "Duplicate flavour names detected"

        return cls(flavours_dict)

    @classmethod
    def from_list(cls, flavours: list[Flavour]) -> FlavourContainer:
        return cls({f.name: f for f in flavours})

    def backgrounds(self, flavour: Flavour, keep_possible_signals: bool = True) -> FlavourContainer:
        bkg = [f for f in self if f.category == flavour.category and f != flavour]
        if not keep_possible_signals:
            bkg = [f for f in bkg if f.name not in {"ujets", "qcd"}]
        return FlavourContainer.from_list(bkg)


Flavours = FlavourContainer.from_yaml()



from __future__ import annotations

from ftag.hdf5.h5reader import H5Reader
from ftag.hdf5.h5utils import cast_dtype, get_dtype, join_structured_arrays, structured_from_dict
from ftag.hdf5.h5writer import H5Writer

__all__ = [
    "H5Reader",
    "H5Writer",
    "cast_dtype",
    "get_dtype",
    "join_structured_arrays",
    "structured_from_dict",
]



from __future__ import annotations

from typing import Callable

import numpy as np

from ftag.flavour import Flavour, Flavours


def discriminant(
    jets: np.ndarray,
    tagger: str,
    signal: Flavour,
    fxs: dict[str, float],
    epsilon: float = 1e-10,
) -> np.ndarray:
    """
    Get the tagging discriminant.

    Calculated as the logarithm of the ratio of a specified signal probability
    to a weighted sum ofbackground probabilities.

    Parameters
    ----------
    jets : np.ndarray
        Structed jet array containing tagger scores.
    tagger : str
        Name of the tagger, used to construct field names.
    signal : str
        Type of signal.
    fxs : dict[str, float]
        Dict of background probability names and their fractions.
        If a fraction is None, it is calculated as (1 - sum of provided fractions).
    epsilon : float, optional
        A small value added to probabilities to prevent division by zero, by default 1e-10.

    Returns
    -------
    np.ndarray
        The tagger discriminant values for the jets.

    Raises
    ------
    ValueError
        If a fraction is specified for a denominator that is not present in the input array.
    """
    denominator = 0.0
    for d, fx in fxs.items():
        name = f"{tagger}_{d}"
        if fx > 0 and name not in jets.dtype.names:
            raise ValueError(f"Nonzero fx for {d}, but '{name}' not found in input array.")
        denominator += jets[name] * fx if name in jets.dtype.names else 0
    return np.log((jets[f"{tagger}_{signal.px}"] + epsilon) / (denominator + epsilon))


def tautag_dicriminant(jets, tagger, fb, fc, epsilon=1e-10):
    fxs = {"pb": fb, "pc": fc, "pu": 1 - fb - fc}
    return discriminant(jets, tagger, Flavours.taujets, fxs, epsilon=epsilon)


def btag_discriminant(jets, tagger, fc, ftau=0, epsilon=1e-10):
    fxs = {"pc": fc, "ptau": ftau, "pu": 1 - fc - ftau}
    return discriminant(jets, tagger, Flavours.bjets, fxs, epsilon=epsilon)


def ctag_discriminant(jets, tagger, fb, ftau=0, epsilon=1e-10):
    fxs = {"pb": fb, "ptau": ftau, "pu": 1 - fb - ftau}
    return discriminant(jets, tagger, Flavours.cjets, fxs, epsilon=epsilon)


def hbb_discriminant(jets, tagger, ftop=0.25, fhcc=0.02, epsilon=1e-10):
    fxs = {"phcc": fhcc, "ptop": ftop, "pqcd": 1 - ftop - fhcc}
    return discriminant(jets, tagger, Flavours.hbb, fxs, epsilon=epsilon)


def hcc_discriminant(jets, tagger, ftop=0.25, fhbb=0.3, epsilon=1e-10):
    fxs = {"phbb": fhbb, "ptop": ftop, "pqcd": 1 - ftop - fhbb}
    return discriminant(jets, tagger, Flavours.hcc, fxs, epsilon=epsilon)


def get_discriminant(
    jets: np.ndarray, tagger: str, signal: Flavour | str, epsilon: float = 1e-10, **fxs
):
    """Calculate the b-tag or c-tag discriminant for a given tagger.

    Parameters
    ----------
    jets : np.ndarray
        Structured array of jets containing tagger outputs
    tagger : str
        Name of the tagger
    signal : Flavour
        Signal flavour (bjets/cjets or hbb/hcc)
    epsilon : float, optional
        Small number to avoid division by zero, by default 1e-10
    **fxs : dict
        Fractions for the different background flavours.

    Returns
    -------
    np.ndarray
        Array of discriminant values.

    Raises
    ------
    ValueError
        If the signal flavour is not recognised.
    """
    tagger_funcs: dict[str, Callable] = {
        "bjets": btag_discriminant,
        "cjets": ctag_discriminant,
        "taujets": tautag_dicriminant,
        "hbb": hbb_discriminant,
        "hcc": hcc_discriminant,
    }

    if str(signal) not in tagger_funcs:
        raise ValueError(f"Signal flavour must be one of {list(tagger_funcs.keys())}, not {signal}")

    func: Callable = tagger_funcs[str(Flavours[signal])]
    return func(jets, tagger, **fxs, epsilon=epsilon)



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Dict

import numpy as np

__all__ = ["Transform"]

Batch = Dict[str, np.ndarray]


@dataclass
class Transform:
    variable_map: dict[str, dict[str, str]] | None = None
    ints_map: dict[str, dict[str, dict[int, int]]] | None = None
    floats_map: dict[str, dict[str, str | Callable]] | None = None

    def __post_init__(self):
        self.variable_map = self.variable_map or {}
        self.variable_map_inv = {
            k: {v: k for k, v in v.items()} for k, v in self.variable_map.items()
        }
        self.ints_map = self.ints_map or {}
        self.floats_map = self.floats_map or {}

        # convert string to numpy function
        for group, map_dict in self.floats_map.items():
            for variable, func in map_dict.items():
                self.floats_map[group][variable] = getattr(np, func)

    def __call__(self, batch: Batch) -> Batch:
        batch = self.map_ints(batch)
        batch = self.map_floats(batch)
        return self.map_variables(batch)

    def map_variables(self, batch: Batch) -> Batch:
        """
        Rename variables in a batch of data.

        Parameters
        ----------
        batch : Batch
            Dict of structured numpy arrays.

        Returns
        -------
        Batch
            Dict of structured numpy arrays with renamed variables.
        """
        assert self.variable_map is not None
        for group in self.variable_map:
            if group in batch:
                batch[group] = batch[group].astype(self.map_dtype(group, batch[group].dtype))
        return batch

    def map_ints(self, batch: Batch) -> Batch:
        """
        Map integer values to new values.

        Parameters
        ----------
        batch : Batch
            Dict of structured numpy arrays.

        Returns
        -------
        Batch
            Dict of structured numpy arrays with mapped integer values.
        """
        assert self.ints_map is not None
        for group, map_dict in self.ints_map.items():
            if group not in batch:
                continue
            for variable, int_map in map_dict.items():
                if variable not in batch[group].dtype.names:
                    continue
                data = batch[group][variable]
                for old, new in int_map.items():
                    data[data == old] = new
        return batch

    def map_floats(self, batch: Batch) -> Batch:
        """
        Transform float values.

        Parameters
        ----------
        batch : Batch
            Dict of structured numpy arrays.

        Returns
        -------
        Batch
            Dict of structured numpy arrays with transformed float values.
        """
        assert self.floats_map is not None
        for group, map_dict in self.floats_map.items():
            if group not in batch:
                continue
            for variable, func in map_dict.items():
                assert callable(func)
                batch[group][variable] = func(batch[group][variable])
        return batch

    def map_dtype(self, name: str, dtype: np.dtype) -> np.dtype:
        assert self.variable_map is not None
        if not (map_dict := self.variable_map.get(name.lstrip("/"))):
            return dtype
        names = list(dtype.names)
        for old, new in map_dict.items():
            if old in names and new in names:
                raise ValueError(f"Variables {old, new} already exists in {name}.")
        return np.dtype([(map_dict.get(name, name), dtype[name]) for name in names])

    def map_variable_names(self, name: str, variables: list[str], inverse=False) -> list[str]:
        variable_map = self.variable_map_inv if inverse else self.variable_map
        if not (map_dict := variable_map.get(name.lstrip("/"))):
            return variables
        return [map_dict.get(name, name) for name in variables]



from __future__ import annotations

import argparse
import glob
from pathlib import Path

import h5py


def parse_args(args):
    parser = argparse.ArgumentParser(
        description="Create a lightweight wrapper around a set of h5 files"
    )
    parser.add_argument("pattern", type=Path, help="quotes-enclosed glob pattern of files to merge")
    parser.add_argument("output", type=Path, help="path to output virtual file")
    return parser.parse_args(args)


def get_virtual_layout(fnames: list[str], group: str):
    # get sources
    sources = []
    total = 0
    for fname in fnames:
        with h5py.File(fname) as f:
            vsource = h5py.VirtualSource(f[group])
            total += vsource.shape[0]
            sources.append(vsource)

    # define layout of the vds
    with h5py.File(fnames[0]) as f:
        dtype = f[group].dtype
        shape = f[group].shape
    shape = (total,) + shape[1:]
    layout = h5py.VirtualLayout(shape=shape, dtype=dtype)

    # fill the vds
    idx = 0
    for source in sources:
        length = source.shape[0]
        layout[idx : idx + length] = source
        idx += length

    return layout


def create_virtual_file(
    pattern: Path | str, out_fname: Path | None = None, overwrite: bool = False
):
    # get list of filenames
    fnames = glob.glob(str(pattern))
    if not fnames:
        raise FileNotFoundError(f"No files matched pattern {pattern}")

    # infer output path if not given
    if out_fname is None:
        assert len({Path(fname).parent for fname in fnames}) == 1
        out_fname = Path(fnames[0]).parent / "vds" / "vds.h5"
    else:
        out_fname = Path(out_fname)

    # check if file already exists
    if not overwrite and out_fname.is_file():
        return out_fname

    # identify common groups across all files
    common_groups: set[str] = set()
    for fname in fnames:
        with h5py.File(fname) as f:
            groups = set(f.keys())
            common_groups = groups if not common_groups else common_groups.intersection(groups)

    if not common_groups:
        raise ValueError("No common groups found across files")

    # create virtual file
    out_fname.parent.mkdir(exist_ok=True)
    with h5py.File(out_fname, "w") as f:
        for group in common_groups:
            layout = get_virtual_layout(fnames, group)
            f.create_virtual_dataset(group, layout)
            attrs_dict: dict = {}
            for fname in fnames:
                with h5py.File(fname) as g:
                    for name, value in g[group].attrs.items():
                        if name not in attrs_dict:
                            attrs_dict[name] = []
                        attrs_dict[name].append(value)
            for name, value in attrs_dict.items():
                if len(value) > 0:
                    f[group].attrs[name] = value[0]

    return out_fname


def main(args=None) -> None:
    args = parse_args(args)
    print(f"Globbing {args.pattern}...")
    create_virtual_file(args.pattern, args.output, overwrite=True)
    with h5py.File(args.output) as f:
        key = next(iter(f.keys()))
        num = len(f[key])
    print(f"Virtual dataset '{key}' has {num:,} entries")
    print(f"Saved virtual file to {args.output.resolve()}")


if __name__ == "__main__":
    main()



"""A script to move (i.e. rename) a dataset in an h5 file."""

from __future__ import annotations

import argparse

import h5py

from ftag.cli_utils import HelpFormatter, valid_path


def parse_args(args):
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=HelpFormatter)
    parser.add_argument("--fname", required=True, type=valid_path, help="path to h5 file")
    parser.add_argument(
        "--src",
        required=True,
        type=str,
        help="path within h5 file to source dataset, starts with '/'",
    )
    parser.add_argument(
        "--dst",
        required=True,
        type=str,
        help="path within h5 file to destination dataset, starts with '/'",
    )
    return parser.parse_args(args)


def main(args=None) -> None:
    args = parse_args(args)
    print(f"Moving {args.src} to {args.dst} in {args.fname}")
    f = h5py.File(args.fname, "a")
    f.move(args.src, args.dst)
    f.close()



from __future__ import annotations

from pathlib import Path
from tempfile import NamedTemporaryFile, mkdtemp

import h5py
import numpy as np
from numpy.lib.recfunctions import unstructured_to_structured as u2s

from ftag.hdf5 import join_structured_arrays

__all__ = ["get_mock_file"]

JET_VARS = [
    ("pt", "f4"),
    ("eta", "f4"),
    ("abs_eta", "f4"),
    ("mass", "f4"),
    ("pt_btagJes", "f4"),
    ("eta_btagJes", "f4"),
    ("n_tracks", "i4"),
    ("HadronConeExclTruthLabelID", "i4"),
    ("HadronConeExclTruthLabelPt", "f4"),
    ("R10TruthLabel_R22v1", "i4"),
    ("n_truth_promptLepton", "i4"),
    ("flavour_label", "i4"),
]

TRACK_VARS = [
    ("d0", "f4"),
    ("z0SinTheta", "f4"),
    ("dphi", "f4"),
    ("deta", "f4"),
    ("qOverP", "f4"),
    ("IP3D_signed_d0_significance", "f4"),
    ("IP3D_signed_z0_significance", "f4"),
    ("phiUncertainty", "f4"),
    ("thetaUncertainty", "f4"),
    ("qOverPUncertainty", "f4"),
    ("numberOfPixelHits", "i4"),
    ("numberOfSCTHits", "i4"),
    ("numberOfInnermostPixelLayerHits", "i4"),
    ("numberOfNextToInnermostPixelLayerHits", "i4"),
    ("numberOfInnermostPixelLayerSharedHits", "i4"),
    ("numberOfInnermostPixelLayerSplitHits", "i4"),
    ("numberOfPixelSharedHits", "i4"),
    ("numberOfPixelSplitHits", "i4"),
    ("numberOfSCTSharedHits", "i4"),
    ("numberOfPixelHoles", "i4"),
    ("numberOfSCTHoles", "i4"),
]


def softmax(x, axis=None):
    e_x = np.exp(x - np.max(x, axis=axis, keepdims=True))
    return e_x / e_x.sum(axis=axis, keepdims=True)


def get_mock_scores(labels: np.ndarray, is_xbb: bool = False):
    means = [
        [2, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 3.5, 0],
        [0, 0, 0, 1],
    ]
    if not is_xbb:
        label_dict = {"u": 0, "c": 4, "b": 5, "tau": 15}
        label_mapping = dict(zip(label_dict.values(), means))
    else:
        label_dict = {"hbb": 11, "hcc": 12, "top": 1, "qcd": 10}
        label_mapping = dict(zip(label_dict.values(), means))

    rng = np.random.default_rng(42)
    nclass = len(label_dict)
    scores = np.zeros((len(labels), nclass))
    scales = [1, 2.5, 5, 1]
    for i, (label, count) in enumerate(zip(*np.unique(labels, return_counts=True))):
        scores[labels == label] = rng.normal(
            loc=label_mapping[label], scale=scales[i], size=(count, nclass)
        )
    scores = softmax(scores, axis=1)
    name = "MockXbbTagger" if is_xbb else "MockTagger"
    cols = [f"{name}_p{x}" for x in label_dict]
    return u2s(scores, dtype=np.dtype([(name, "f4") for name in cols]))


def mock_jets(num_jets=1000) -> np.ndarray:
    # setup jets
    rng = np.random.default_rng(42)
    jets_dtype = np.dtype(JET_VARS)
    jets = u2s(rng.random((num_jets, len(JET_VARS))), jets_dtype)
    jets["flavour_label"] = rng.choice([0, 4, 5], size=num_jets)
    jets["pt"] *= 400e3
    jets["mass"] *= 50e3
    jets["eta"] = (jets["eta"] - 0.5) * 6.0
    jets["abs_eta"] = np.abs(jets["eta"])
    jets["n_truth_promptLepton"] = 0

    # add tagger scores
    jets["HadronConeExclTruthLabelID"] = rng.choice([0, 4, 5, 15], size=num_jets)
    jets["R10TruthLabel_R22v1"] = rng.choice([1, 10, 11, 12], size=num_jets)
    scores = get_mock_scores(jets["HadronConeExclTruthLabelID"])
    xbb_scores = get_mock_scores(jets["R10TruthLabel_R22v1"], is_xbb=True)
    return join_structured_arrays([jets, scores, xbb_scores])


def mock_tracks(num_jets=1000, num_tracks=40) -> np.ndarray:
    rng = np.random.default_rng(42)
    tracks_dtype = np.dtype(TRACK_VARS)
    tracks = u2s(rng.random((num_jets, num_tracks, len(TRACK_VARS))), tracks_dtype)
    tracks["d0"] *= 5
    valid = rng.choice([True, False], size=(num_jets, num_tracks))
    valid = valid.astype(bool).view(dtype=np.dtype([("valid", bool)]))
    return join_structured_arrays([tracks, valid])


def get_mock_file(
    num_jets=1000,
    fname: str | None = None,
    tracks_name: str = "tracks",
    num_tracks: int = 40,
) -> tuple[str, h5py.File]:
    jets = mock_jets(num_jets)

    # create a tempfile in a new folder
    if fname is None:
        fname = NamedTemporaryFile(suffix=".h5", dir=mkdtemp()).name
    else:
        Path(fname).parent.mkdir(exist_ok=True, parents=True)
    f = h5py.File(fname, "w")
    f.create_dataset("jets", data=jets)
    f.attrs["test"] = "test"
    f["jets"].attrs["test"] = "test"

    # setup tracks
    if tracks_name:
        tracks = mock_tracks(num_jets, num_tracks)
        f.create_dataset(tracks_name, data=tracks)

    return fname, f



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import h5py
import numpy as np

import ftag


@dataclass
class H5Writer:
    """Writes jets to an HDF5 file.

    Parameters
    ----------
    dst : Path | str
        Path to the output file.
    dtypes : dict[str, np.dtype]
        Dictionary of group names and their corresponding dtypes.
    num_jets : int
        Number of jets to write.
    shapes : dict[str, int], optional
        Dictionary of group names and their corresponding shapes.
    jets_name : str, optional
        Name of the jets group. Default is "jets".
    add_flavour_label : bool, optional
        Whether to add a flavour label to the jets group. Default is False.
    compression : str, optional
        Compression algorithm to use. Default is "lzf".
    precision : str | None, optional
        Precision to use. Default is None.
    shuffle : bool, optional
        Whether to shuffle the jets before writing. Default is True.
    """

    dst: Path | str
    dtypes: dict[str, np.dtype]
    shapes: dict[str, tuple[int, ...]]
    jets_name: str = "jets"
    add_flavour_label: bool = False
    compression: str = "lzf"
    precision: str = "full"
    shuffle: bool = True

    def __post_init__(self):
        self.num_written = 0
        self.rng = np.random.default_rng(42)
        self.num_jets = [shape[0] for shape in self.shapes.values()]
        assert len(set(self.num_jets)) == 1, "Must have same number of jets per group"
        self.num_jets = self.num_jets[0]

        if self.precision == "full":
            self.fp_dtype = np.float32
        elif self.precision == "half":
            self.fp_dtype = np.float16
        else:
            raise ValueError(f"Invalid precision: {self.precision}")

        self.dst = Path(self.dst)
        self.dst.parent.mkdir(parents=True, exist_ok=True)
        self.file = h5py.File(self.dst, "w")
        self.add_attr("writer_version", ftag.__version__)

        for name, dtype in self.dtypes.items():
            self.create_ds(name, dtype)

    @classmethod
    def from_file(cls, source: Path, num_jets: int | None = None, **kwargs) -> H5Writer:
        with h5py.File(source, "r") as f:
            dtypes = {name: ds.dtype for name, ds in f.items()}
            shapes = {name: ds.shape for name, ds in f.items()}
            if num_jets is not None:
                shapes = {name: (num_jets,) + shape[1:] for name, shape in shapes.items()}
            compression = [ds.compression for ds in f.values()]
            assert len(set(compression)) == 1, "Must have same compression for all groups"
            compression = compression[0]
            if compression not in kwargs:
                kwargs["compression"] = compression
        return cls(dtypes=dtypes, shapes=shapes, **kwargs)

    def create_ds(self, name: str, dtype: np.dtype) -> None:
        if name == self.jets_name and self.add_flavour_label:
            dtype = np.dtype([*dtype.descr, ("flavour_label", "i4")])

        # adjust dtype based on specified precision
        dtype = np.dtype([
            (field, self.fp_dtype if np.issubdtype(dt, np.floating) else dt)
            for field, dt in dtype.descr
        ])

        # optimal chunking is around 100 jets, only aply for track groups
        shape = self.shapes[name]
        chunks = (100,) + shape[1:] if shape[1:] else None

        # note: enabling the hd5 shuffle filter doesn't improve write performance
        self.file.create_dataset(
            name, dtype=dtype, shape=shape, compression=self.compression, chunks=chunks
        )

    def close(self) -> None:
        with h5py.File(self.dst) as f:
            written = len(f[self.jets_name])
        if self.num_written != written:
            raise ValueError(
                f"Attemped to close file {self.dst} when only {self.num_written:,} out of"
                f" {written:,} jets have been written"
            )
        self.file.close()

    def get_attr(self, name, group=None):
        with h5py.File(self.dst) as f:
            obj = f[group] if group else f
            return obj.attrs[name]

    def add_attr(self, name, data, group=None) -> None:
        obj = self.file[group] if group else self.file
        obj.attrs.create(name, data)

    def copy_attrs(self, fname: Path) -> None:
        with h5py.File(fname) as f:
            for name, value in f.attrs.items():
                self.add_attr(name, value)
            for name, ds in f.items():
                for attr_name, value in ds.attrs.items():
                    self.add_attr(attr_name, value, group=name)

    def write(self, data: dict[str, np.array]) -> None:
        if (total := self.num_written + len(data[self.jets_name])) > self.num_jets:
            raise ValueError(
                f"Attempted to write more jets than expected: {total:,} > {self.num_jets:,}"
            )
        idx = np.arange(len(data[self.jets_name]))
        if self.shuffle:
            self.rng.shuffle(idx)
            data = {name: array[idx] for name, array in data.items()}

        low = self.num_written
        high = low + len(idx)
        for group in self.dtypes:
            self.file[group][low:high] = data[group]
        self.num_written += len(idx)



"""Calculate tagger working points."""

from __future__ import annotations

import argparse
from pathlib import Path

import numpy as np
import yaml

from ftag.cli_utils import HelpFormatter
from ftag.cuts import Cuts
from ftag.flavour import Flavours
from ftag.hdf5 import H5Reader
from ftag.wps.discriminant import get_discriminant


def parse_args(args):
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=HelpFormatter,
    )
    parser.add_argument(
        "--ttbar",
        required=True,
        type=Path,
        help="path to ttbar sample (supports globbing)",
    )
    parser.add_argument(
        "--zprime",
        required=False,
        type=Path,
        help="path to zprime (supports globbing). WPs from ttbar will be reused for zprime",
    )
    parser.add_argument(
        "-e",
        "--effs",
        nargs="+",
        type=float,
        help="efficiency working point(s). If -r is specified, values should be 1/efficiency",
    )
    parser.add_argument(
        "-t",
        "--tagger",
        nargs="+",
        required=True,
        type=str,
        help="tagger name(s)",
    )
    parser.add_argument(
        "-s",
        "--signal",
        default="bjets",
        choices=["bjets", "cjets", "hbb", "hcc"],
        type=str,
        help='signal flavour ("bjets" or "cjets" for b-tagging, "hbb" or "hcc" for Xbb)',
    )
    parser.add_argument(
        "-r",
        "--rejection",
        default=None,
        choices=["ujets", "cjets", "bjets", "hbb", "hcc", "top", "qcd"],
        help="use rejection of specified background class to determine working points",
    )
    parser.add_argument(
        "-d",
        "--disc_cuts",
        nargs="+",
        type=float,
        help="D_x value(s) to calculate efficiency at",
    )
    parser.add_argument(
        "-n",
        "--num_jets",
        default=1_000_000,
        type=int,
        help="use this many jets (post selection)",
    )
    parser.add_argument(
        "--ttbar_cuts",
        nargs="+",
        default=["pt > 20e3"],
        type=list,
        help="selection to apply to ttbar (|eta| < 2.5 is always applied)",
    )
    parser.add_argument(
        "--zprime_cuts",
        nargs="+",
        default=["pt > 250e3"],
        type=list,
        help="selection to apply to zprime (|eta| < 2.5 is always applied)",
    )
    parser.add_argument(
        "-o",
        "--outfile",
        type=Path,
        help="save results to yaml instead of printing",
    )
    parser.add_argument(
        "--xbb",
        action="store_true",
        help="Enable Xbb tagging which expects two fx values ftop and fhcc/fhbb for each tagger",
    )
    parser.add_argument(
        "--fb",
        nargs="+",
        type=float,
        help="fb value(s) for each tagger",
    )
    parser.add_argument(
        "--fc",
        nargs="+",
        type=float,
        help="fc value(s) for each tagger",
    )
    parser.add_argument(
        "--ftau",
        nargs="+",
        type=float,
        help="ftau value(s) for each tagger",
    )
    parser.add_argument(
        "--ftop",
        nargs="+",
        type=float,
        help="ftop value(s) for each tagger",
    )
    parser.add_argument(
        "--fhbb",
        nargs="+",
        type=float,
        help="fhbb value(s) for each tagger",
    )
    parser.add_argument(
        "--fhcc",
        nargs="+",
        type=float,
        help="fhcc value(s) for each tagger",
    )
    args = parser.parse_args(args)

    args.signal = Flavours[args.signal]

    if args.effs and args.disc_cuts:
        raise ValueError("Cannot specify both --effs and --disc_cuts")
    if not args.effs and not args.disc_cuts:
        raise ValueError("Must specify either --effs or --disc_cuts")

    if args.xbb:
        if args.signal not in {Flavours.hbb, Flavours.hcc}:
            raise ValueError("Xbb tagging only supports hbb or hcc signal flavours")
        if args.fb or args.fc or args.ftau:
            raise ValueError("For Xbb tagging, fb, fc and ftau should not be specified")
        if not args.ftop:
            raise ValueError("For Xbb tagging, ftop should be specified")
        if args.signal == "hbb" and not args.fhcc:
            raise ValueError("For hbb tagging, fhcc should be specified")
        if args.signal == "hcc" and not args.fhbb:
            raise ValueError("For hcc tagging, fhbb should be specified")
    else:
        if args.ftop or args.fhbb or args.fhcc:
            raise ValueError("For single-b tagging, ftop, fhbb and fhcc should not be specified")
        if args.signal == "bjets" and not args.fc:
            raise ValueError("For bjets tagging, fc should be specified")
        if args.signal == "cjets" and not args.fb:
            raise ValueError("For cjets tagging, fb should be specified")
        if args.ftau is None:
            args.ftau = [0.0] * len(args.tagger)

    for fx in ["fb", "fc", "ftau", "ftop", "fhbb", "fhcc"]:
        if getattr(args, fx) and len(getattr(args, fx)) != len(args.tagger):
            raise ValueError(f"Number of {fx} values must match number of taggers")

    return args


def get_fxs_from_args(args):
    if args.signal == Flavours.bjets:
        fxs = {"fc": args.fc, "ftau": args.ftau}
    elif args.signal == Flavours.cjets:
        fxs = {"fb": args.fb, "ftau": args.ftau}
    elif args.signal == Flavours.hbb:
        fxs = {"ftop": args.ftop, "fhcc": args.fhcc}
    elif args.signal == Flavours.hcc:
        fxs = {"ftop": args.ftop, "fhbb": args.fhbb}
    assert fxs is not None
    return [{k: v[i] for k, v in fxs.items()} for i in range(len(args.tagger))]


def get_eff_rej(jets, disc, wp, flavs):
    out = {"eff": {}, "rej": {}}
    for bkg in list(flavs):
        bkg_disc = disc[bkg.cuts(jets).idx]
        eff = sum(bkg_disc > wp) / len(bkg_disc)
        out["eff"][str(bkg)] = float(f"{eff:.3g}")
        out["rej"][str(bkg)] = float(f"{1 / eff:.3g}")
    return out


def get_rej_eff_at_disc(jets, tagger, signal, disc_cuts, **fxs):
    disc = get_discriminant(jets, tagger, signal, **fxs)
    d = {}
    flavs = Flavours.by_category("single-btag")
    for dcut in disc_cuts:
        d[str(dcut)] = {"eff": {}, "rej": {}}
        for f in flavs:
            e_discs = disc[f.cuts(jets).idx]
            eff = sum(e_discs > dcut) / len(e_discs)
            d[str(dcut)]["eff"][str(f)] = float(f"{eff:.3g}")
            d[str(dcut)]["rej"][str(f)] = 1 / float(f"{eff:.3g}")
    return d


def setup_common_parts(args):
    flavs = Flavours.by_category("single-btag") if not args.xbb else Flavours.by_category("xbb")
    default_cuts = Cuts.from_list(["eta > -2.5", "eta < 2.5"])
    ttbar_cuts = Cuts.from_list(args.ttbar_cuts) + default_cuts
    zprime_cuts = Cuts.from_list(args.zprime_cuts) + default_cuts

    # prepare to load jets
    all_vars = next(iter(flavs)).cuts.variables
    reader = H5Reader(args.ttbar)
    jet_vars = reader.dtypes()["jets"].names
    for tagger in args.tagger:
        all_vars += [f"{tagger}_{f.px}" for f in flavs if (f"{tagger}_{f.px}" in jet_vars)]

    # load jets
    jets = reader.load({"jets": all_vars}, args.num_jets, cuts=ttbar_cuts)["jets"]
    zp_jets = None
    if args.zprime:
        zp_reader = H5Reader(args.zprime)
        zp_jets = zp_reader.load({"jets": all_vars}, args.num_jets, cuts=zprime_cuts)["jets"]

    return jets, zp_jets, flavs


def get_working_points(args=None):
    jets, zp_jets, flavs = setup_common_parts(args)
    fxs = get_fxs_from_args(args)

    # loop over taggers
    out = {}
    for i, tagger in enumerate(args.tagger):
        # calculate discriminant
        out[tagger] = {"signal": str(args.signal), **fxs[i]}
        disc = get_discriminant(jets, tagger, args.signal, **fxs[i])

        # loop over efficiency working points
        for eff in args.effs:
            d = out[tagger][f"{eff:.0f}"] = {}

            wp_flavour = args.signal
            if args.rejection:
                eff = 100 / eff  # noqa: PLW2901
                wp_flavour = args.rejection

            wp_disc = disc[flavs[wp_flavour].cuts(jets).idx]
            wp = d["cut_value"] = round(float(np.percentile(wp_disc, 100 - eff)), 3)

            # calculate eff and rej for each flavour
            d["ttbar"] = get_eff_rej(jets, disc, wp, flavs)

            # calculate for zprime
            if args.zprime:
                zp_disc = get_discriminant(zp_jets, tagger, Flavours[args.signal], **fxs[i])
                d["zprime"] = get_eff_rej(zp_jets, zp_disc, wp, flavs)

    if args.outfile:
        with open(args.outfile, "w") as f:
            yaml.dump(out, f, sort_keys=False)
            return None
    else:
        return out


def get_efficiencies(args=None):
    jets, zp_jets, _ = setup_common_parts(args)
    fxs = get_fxs_from_args(args)

    out = {}
    for i, tagger in enumerate(args.tagger):
        out[tagger] = {"signal": str(args.signal), **fxs[i]}

        out[tagger]["ttbar"] = get_rej_eff_at_disc(
            jets, tagger, args.signal, args.disc_cuts, **fxs[i]
        )
        if args.zprime:
            out[tagger]["zprime"] = get_rej_eff_at_disc(
                zp_jets, tagger, args.signal, args.disc_cuts, **fxs[i]
            )

    if args.outfile:
        with open(args.outfile, "w") as f:
            yaml.dump(out, f, sort_keys=False)
            return None
    else:
        return out


def main(args=None):
    args = parse_args(args)

    if args.effs:
        out = get_working_points(args)
    elif args.disc_cuts:
        out = get_efficiencies(args)

    if out:
        print(yaml.dump(out, sort_keys=False))
        return out

    return None


if __name__ == "__main__":
    main()



"""A script to split a large h5 file into smaller h5 files along the first index."""

from __future__ import annotations

import argparse
from pathlib import Path

import h5py

from ftag.cli_utils import HelpFormatter
from ftag.hdf5 import H5Reader, H5Writer


def parse_args(args):
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=HelpFormatter)
    parser.add_argument("--src", required=True, type=Path, help="path to source h5 file")
    parser.add_argument(
        "--dst",
        type=Path,
        help=(
            "output directory contaning split files. by default use a new directory in the same"
            " directory as the source file"
        ),
    )
    parser.add_argument(
        "-n", "--jets_per_file", type=int, default=1_000_000, help="number of jets per output file"
    )
    parser.add_argument(
        "-b",
        "--batch_size",
        type=int,
        default=100_000,
        help="number of jets to read/write at a time",
    )
    return parser.parse_args(args)


def main(args=None):
    args = parse_args(args)

    src = args.src
    dst = args.dst
    if dst is None:
        dst = src.parent / f"split_{src.stem}"
    jets_per_file = args.jets_per_file

    print(f"\nSplitting: {src}")
    print(f"Destination: {dst}")
    with h5py.File(src, "r") as f:
        total_jets = next(iter(f.values())).shape[0]

    num_full_files = total_jets // jets_per_file
    remainder = total_jets % jets_per_file
    num_total_files = num_full_files + (1 if remainder != 0 else 0)
    print(f"\n{total_jets:,} jets will be split across {num_total_files:,} files\n")

    reader = H5Reader(src, batch_size=args.batch_size, shuffle=False)
    variables = dict.fromkeys(reader.dtypes().keys())
    for i in range(num_total_files):
        start = i * jets_per_file
        out = dst / f"{src.stem}-split_{i}.h5"
        num = jets_per_file if i < num_full_files else remainder
        writer = H5Writer.from_file(src, dst=out, num_jets=num, shuffle=False)
        for batch in reader.stream(variables=variables, num_jets=num, start=start):
            writer.write(batch)
            total_written = start + writer.num_written
            pct_done = total_written / total_jets
            print(f"\rProcessed {total_written:,}/{total_jets:,} jets ({pct_done:.1%})", end="")
        writer.copy_attrs(src)
        writer.close()
    print("\nDone!\n")


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass

import numpy as np

from ftag import Flavours
from ftag.flavour import Flavour, FlavourContainer
from ftag.hdf5 import join_structured_arrays, structured_from_dict


@dataclass
class Labeller:
    """
    Defines a labelling scheme.

    Labels are [0, ..., n] and are assigned using pre-defined selections.

    Parameters
    ----------
    labels : FlavourContainer | list[str | Flavour]
        The labels to be use.
    require_labels : bool
        Whether to require that all objects are labelled.
    """

    labels: FlavourContainer | list[str | Flavour]
    require_labels: bool = True

    def __post_init__(self) -> None:
        if isinstance(self.labels, FlavourContainer):
            self.labels = list(self.labels)
        self.labels = sorted([Flavours[label] for label in self.labels])

    def get_labels(self, array: np.ndarray) -> np.ndarray:
        """
        Returns the labels for the given array.

        Parameters
        ----------
        array : np.ndarray
            The array to label.

        Returns
        -------
        np.ndarray
            The labels for the given array.

        Raises
        ------
        ValueError
            If the `require_labels` attribute is set to `True` and some objects were not labelled.
        """
        labels = -1 * np.ones_like(array, dtype=int)
        for i, label in enumerate(self.labels):
            labels[label.cuts(array).idx] = i

        if self.require_labels and -1 in labels:
            raise ValueError("Some objects were not labelled")

        return labels[labels != -1]

    def add_labels(self, array: np.ndarray, label_name: str = "labels") -> np.ndarray:
        """
        Adds the labels to the given array.

        Parameters
        ----------
        array : np.ndarray
            The array to label.
        label_name : str
            The name of the label column.

        Returns
        -------
        np.ndarray
            The array with the labels added.

        Raises
        ------
        ValueError
            If the `require_labels` attribute is set to `False`.
        """
        if not self.require_labels:
            raise ValueError("Cannot add labels if require_labels is set to False")
        labels = self.get_labels(array)
        labels = structured_from_dict({label_name: labels})
        return join_structured_arrays([array, labels])



from __future__ import annotations

import subprocess
import sys
from subprocess import CalledProcessError


class GitError(Exception):
    pass


def is_git_repo(path) -> bool:
    try:
        subprocess.check_output(["git", "rev-parse", "--is-inside-work-tree", "HEAD"], cwd=path)
    except CalledProcessError:
        return False
    else:
        return True


def check_for_uncommitted_changes(path) -> None:
    if not is_git_repo(path):
        return
    if "pytest" in sys.modules:
        return

    try:
        subprocess.check_output(["git", "diff", "--quiet", "--exit-code"], cwd=path)
    except CalledProcessError:
        raise GitError(
            "Uncommitted changes detected. Please commit them before running, or use --force."
        ) from None


def check_for_fork(path, upstream) -> None:
    cmd = ["git", "remote", "get-url", "origin"]
    if not is_git_repo(path):
        return

    origin = subprocess.check_output(cmd, cwd=path).decode("utf-8").strip()
    if upstream not in origin:
        raise GitError(f"Your origin {origin} is not a fork of the upstream repo {upstream}")


def create_and_push_tag(path, upstream, tagname, msg) -> None:
    print(f"Pushing tag {tagname}")
    if not is_git_repo(path):
        return
    check_for_fork(path, upstream)
    subprocess.check_output(["git", "tag", tagname, "-m", msg], cwd=path)
    subprocess.check_output(["git", "push", "-q", "origin", "--tags"], cwd=path)


def get_git_hash(path):
    if not is_git_repo(path):
        return None

    git_hash = subprocess.check_output(["git", "rev-parse", "--short", "HEAD"], cwd=path)
    return git_hash.decode("ascii").strip()



from __future__ import annotations

import logging as log
import math
from collections.abc import Generator
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path

import h5py
import numpy as np

from ftag.cuts import Cuts
from ftag.hdf5.h5utils import get_dtype
from ftag.sample import Sample
from ftag.transform import Transform


@dataclass
class H5SingleReader:
    fname: Path | str
    batch_size: int = 100_000
    jets_name: str = "jets"
    precision: str | None = None
    shuffle: bool = True
    do_remove_inf: bool = False
    transform: Transform | None = None

    def __post_init__(self) -> None:
        self.rng = np.random.default_rng(42)
        self.sample = Sample(self.fname)
        fname = self.sample.virtual_file()
        if len(fname) != 1:
            raise ValueError("H5SingleReader should only read a single file")
        self.fname = fname[0]

    @cached_property
    def num_jets(self) -> int:
        with h5py.File(self.fname) as f:
            return len(f[self.jets_name])

    def get_attr(self, name, group=None):
        with h5py.File(self.fname) as f:
            obj = f[group] if group else f
            return obj.attrs[name]

    def empty(self, ds: h5py.Dataset, variables: list[str]) -> np.ndarray:
        return np.array(0, dtype=get_dtype(ds, variables, self.precision, transform=self.transform))

    def read_chunk(self, ds: h5py.Dataset, array: np.ndarray, low: int) -> np.ndarray:
        high = min(low + self.batch_size, self.num_jets)
        shape = (high - low,) + ds.shape[1:]
        array.resize(shape, refcheck=False)
        ds.read_direct(array, np.s_[low:high])
        return array

    def remove_inf(self, data: dict) -> dict:
        keep_idx = np.full(len(data[self.jets_name]), True)
        for name, array in data.items():
            for var in array.dtype.names:
                isinf = np.isinf(array[var])
                isinf = isinf if name == self.jets_name else isinf.any(axis=-1)
                keep_idx &= ~isinf
                if num_inf := isinf.sum():
                    log.warning(
                        f"{num_inf} inf values detected for variable {var} in"
                        f" {name} array. Removing the affected jets."
                    )
        return {name: array[keep_idx] for name, array in data.items()}

    def stream(
        self,
        variables: dict | None = None,
        num_jets: int | None = None,
        cuts: Cuts | None = None,
        start: int = 0,
    ) -> Generator:
        if num_jets is None:
            num_jets = self.num_jets

        if num_jets > self.num_jets:
            log.warning(
                f"{num_jets:,} jets requested but only {self.num_jets:,} available in {self.fname}."
                " Set to maximum available number!"
            )
            num_jets = self.num_jets

        if variables is None:
            variables = {self.jets_name: None}

        total = 0
        with h5py.File(self.fname) as f:
            arrays = {name: self.empty(f[name], var) for name, var in variables.items()}
            data = {name: self.empty(f[name], var) for name, var in variables.items()}

            # get indices
            indices = list(range(start, self.num_jets + start, self.batch_size))
            if self.shuffle:
                self.rng.shuffle(indices)

            # loop over batches and read file
            for low in indices:
                for name in variables:
                    data[name] = self.read_chunk(f[name], arrays[name], low)

                # apply selections
                if cuts:
                    idx = cuts(data[self.jets_name]).idx
                    data = {name: array[idx] for name, array in data.items()}

                # check for inf and remove
                if self.do_remove_inf:
                    data = self.remove_inf(data)

                # apply transform
                if self.transform:
                    data = self.transform(data)

                # check for completion
                total += len(data[self.jets_name])
                if total >= num_jets:
                    keep = num_jets - (total - len(data[self.jets_name]))
                    data = {name: array[:keep] for name, array in data.items()}
                    yield data
                    break

                yield data


@dataclass
class H5Reader:
    """Reads data from multiple HDF5 files.

    Parameters
    ----------
    fname : Path | str | list[Path | str]
        Path to the HDF5 file or list of paths
    batch_size : int, optional
        Number of jets to read at a time, by default 100_000
    jets_name : str, optional
        Name of the jets dataset, by default "jets"
    precision : str | None, optional
        Cast floats to given precision, by default None
    shuffle : bool, optional
        Read batches in a shuffled order, by default True
    weights : list[float] | None, optional
        Weights for different input datasets, by default None
    do_remove_inf : bool, optional
        Remove jets with inf values, by default False
    transform : Transform | None, optional
        Transform to apply to data, by default None
    equal_jets : bool, optional
        Take the same number of jets (weighted) from each sample, by default True
        If False, use all jets in each sample.
    """

    fname: Path | str | list[Path | str]
    batch_size: int = 100_000
    jets_name: str = "jets"
    precision: str | None = None
    shuffle: bool = True
    weights: list[float] | None = None
    do_remove_inf: bool = False
    transform: Transform | None = None
    equal_jets: bool = True

    def __post_init__(self) -> None:
        self.rng = np.random.default_rng(42)
        if not self.equal_jets:
            log.warning(
                "equal_jets is set to False, which will result in different number of jets taken"
                " from each sample. Be aware that this can affect the resampling, so make sure you"
                " know what you are doing."
            )

        if isinstance(self.fname, (str, Path)):
            self.fname = [self.fname]

        # calculate batch sizes
        if self.weights is None:
            self.weights = [1 / len(self.fname)] * len(self.fname)
        self.batch_sizes = [int(w * self.batch_size) for w in self.weights]

        # create readers
        self.readers = [
            H5SingleReader(
                f,
                b,
                self.jets_name,
                self.precision,
                self.shuffle,
                self.do_remove_inf,
                self.transform,
            )
            for f, b in zip(self.fname, self.batch_sizes)
        ]

    @property
    def num_jets(self) -> int:
        return sum(r.num_jets for r in self.readers)

    @property
    def files(self) -> list[Path]:
        return [Path(r.fname) for r in self.readers]

    def dtypes(self, variables: dict[str, list[str]] | None = None) -> dict[str, np.dtype]:
        dtypes = {}
        with h5py.File(self.files[0]) as f:
            if variables is None:
                for key in f:
                    dtype = f[key].dtype
                    if self.transform:
                        dtype = self.transform.map_dtype(key, dtype)
                    dtypes[key] = dtype
            else:
                for name, var in variables.items():
                    ds = f[name]
                    dtype = get_dtype(ds, var, self.precision, transform=self.transform)
                    dtypes[name] = dtype
        return dtypes

    def shapes(self, num_jets: int, groups: list[str] | None = None) -> dict[str, tuple[int, ...]]:
        if groups is None:
            groups = [self.jets_name]
        shapes = {}
        with h5py.File(self.files[0]) as f:
            for group in groups:
                shape = f[group].shape
                shapes[group] = (num_jets,) + shape[1:]
        return shapes

    def stream(
        self,
        variables: dict | None = None,
        num_jets: int | None = None,
        cuts: Cuts | None = None,
        start: int = 0,
    ) -> Generator:
        """Generate batches of selected jets.

        Parameters
        ----------
        variables : dict | None, optional
            Dictionary of variables to for each group, by default use all jet variables.
        num_jets : int | None, optional
            Total number of selected jets to generate, by default all.
        cuts : Cuts | None, optional
            Selection cuts to apply, by default None
        start : int, optional
            Starting index of the first jet to read, by default 0

        Yields
        ------
        Generator
            Generator of batches of selected jets.
        """
        # Check if number of jets is given, if not, set to maximum available
        if num_jets is None:
            num_jets = self.num_jets

        # Check if variables if given, if not, set to all
        if variables is None:
            variables = {self.jets_name: None}

        if self.jets_name not in variables or variables[self.jets_name] is not None:
            jet_vars = variables.get(self.jets_name, [])
            variables[self.jets_name] = list(jet_vars) + (cuts.variables if cuts else [])

        # get streams for selected jets from each reader
        streams = [
            r.stream(variables, int(r.num_jets / self.num_jets * num_jets), cuts, start)
            for r in self.readers
        ]

        while True:
            samples = []
            # Track which streams have been exhausted
            streams_done = [False] * len(streams)

            # for each unexhausted stream, get the next sample
            for i, stream in enumerate(streams):
                if not streams_done[i]:
                    try:
                        samples.append(next(stream))

                    # if equal_jets is True, we can stop when any stream is done
                    # otherwise if sample is exhausted, mark it as done
                    except StopIteration:
                        if self.equal_jets:
                            return
                        streams_done[i] = True

                # if equal_jets is False, we need to keep going until all streams are done
                if all(streams_done):
                    return

            # combine samples and shuffle
            data = {name: np.concatenate([s[name] for s in samples]) for name in variables}
            if self.shuffle:
                idx = np.arange(len(data[self.jets_name]))
                self.rng.shuffle(idx)
                data = {name: array[idx] for name, array in data.items()}

            # yield batch
            yield data

    def load(
        self, variables: dict | None = None, num_jets: int | None = None, cuts: Cuts | None = None
    ) -> dict:
        """Load multiple batches of selected jets into memory.

        Parameters
        ----------
        variables : dict | None, optional
            Dictionary of variables to for each group, by default use all jet variables.
        num_jets : int | None, optional
            Total number of selected jets to load, by default all.
        cuts : Cuts | None, optional
            Selection cuts to apply, by default None

        Returns
        -------
        dict
            Dictionary of arrays for each group.
        """
        # handle default arguments
        if num_jets == -1:
            num_jets = self.num_jets
        if variables is None:
            variables = {self.jets_name: None}

        # get data from each sample
        data: dict[str, list] = {name: [] for name in variables}
        for batch in self.stream(variables, num_jets, cuts):
            for name, array in batch.items():
                if name in data:
                    data[name].append(array)

        # concatenate batches
        return {name: np.concatenate(array) for name, array in data.items()}

    def estimate_available_jets(self, cuts: Cuts, num: int = 1_000_000) -> int:
        """Estimate the number of jets available after selection cuts.

        Parameters
        ----------
        cuts : Cuts
            Selection cuts to apply.
        num : int, optional
            Number of jets to use for the estimation, by default 1_000_000.

        Returns
        -------
        int
            Estimated number of jets available after selection cuts, rounded down.
        """
        # reset rngs to ensure same jets are used for each sample
        self.rng = np.random.default_rng(42)
        for r in self.readers:
            r.rng = np.random.default_rng(42)

        # if equal jets is True, available jets is based on the smallest sample
        if self.equal_jets:
            num_jets = []
            for r in self.readers:
                stream = r.stream({self.jets_name: cuts.variables}, num)
                all_jets = np.concatenate([batch[self.jets_name].copy() for batch in stream])
                frac_selected = len(cuts(all_jets).values) / len(all_jets)
                num_jets.append(frac_selected * r.num_jets)
            estimated_num_jets = min(num_jets) * len(self.readers)
        # otherwise, available jets is based on all samples
        else:
            all_jets = self.load({self.jets_name: cuts.variables}, num)[self.jets_name]
            frac_selected = len(cuts(all_jets).values) / len(all_jets)
            estimated_num_jets = frac_selected * self.num_jets
        return math.floor(estimated_num_jets * 0.99)



"""atlas-ftag-tools - Common tools for ATLAS flavour tagging software."""

from __future__ import annotations

__version__ = "v0.2.3"


from ftag import hdf5
from ftag.cuts import Cuts
from ftag.flavour import Flavour, Flavours
from ftag.mock import get_mock_file
from ftag.sample import Sample
from ftag.transform import Transform
from ftag.wps.discriminant import get_discriminant
from ftag.wps.working_points import get_working_points

__all__ = [
    "Cuts",
    "Flavour",
    "Flavours",
    "Sample",
    "Transform",
    "__version__",
    "get_discriminant",
    "get_mock_file",
    "get_working_points",
    "hdf5",
]



from __future__ import annotations

import numpy as np
from numpy.lib.recfunctions import unstructured_to_structured as u2s

from ftag.transform import Transform

__all__ = ["cast_dtype", "get_dtype", "join_structured_arrays"]


def get_dtype(
    ds,
    variables: list[str] | None = None,
    precision: str | None = None,
    transform: Transform | None = None,
) -> np.dtype:
    """Return a dtype based on an existing dataset and requested variables.

    Parameters
    ----------
    ds : h5py.Dataset
        Input h5 dataset
    variables : list[str] | None, optional
        List of variables to include in dtype, by default None
    precision : str | None, optional
        Precision to cast floats to, "half" or "full", by default None
    transform : Transform | None, optional
        Transform to apply to variables names, by default None

    Returns
    -------
    np.dtype
        Output dtype

    Raises
    ------
    ValueError
        If variables are not found in dataset
    """
    if variables is None:
        variables = ds.dtype.names

    if (missing := set(variables) - set(ds.dtype.names)) and transform is not None:
        variables = transform.map_variable_names(ds.name, variables, inverse=True)
        missing = set(variables) - set(ds.dtype.names)
    if missing:
        raise ValueError(
            f"Variables {missing} were not found in dataset {ds.name} in file {ds.file.filename}"
        )

    dtype = [(n, x) for n, x in ds.dtype.descr if n in variables]
    if precision:
        dtype = [(n, cast_dtype(x, precision)) for n, x in dtype]

    return np.dtype(dtype)


def cast_dtype(typestr: str, precision: str) -> np.dtype:
    """Cast float type to half or full precision.

    Parameters
    ----------
    typestr : str
        Input type string
    precision : str
        Precision to cast to, "half" or "full"

    Returns
    -------
    np.dtype
        Output dtype

    Raises
    ------
    ValueError
        If precision is not "half" or "full"
    """
    t = np.dtype(typestr)
    if t.kind != "f":
        return t
    if precision == "half":
        return np.dtype("f2")
    if precision == "full":
        return np.dtype("f4")
    raise ValueError(f"Invalid precision {precision}")


def join_structured_arrays(arrays: list):
    """Join a list of structured numpy arrays.

    See https://github.com/numpy/numpy/issues/7811

    Parameters
    ----------
    arrays : list
        List of structured numpy arrays to join

    Returns
    -------
    np.array
        A merged structured array
    """
    dtype: list = sum((a.dtype.descr for a in arrays), [])
    newrecarray = np.empty(arrays[0].shape, dtype=dtype)
    for a in arrays:
        for name in a.dtype.names:
            newrecarray[name] = a[name]

    return newrecarray


def structured_from_dict(d: dict[str, np.ndarray]) -> np.ndarray:
    """Convert a dict to a structured array.

    Parameters
    ----------
    d : dict
        Input dict of numpy arrays

    Returns
    -------
    np.ndarray
        Structured array
    """
    arrays = np.column_stack(list(d.values()))
    dtypes = np.dtype([(k, v.dtype) for k, v in d.items()])
    return u2s(arrays, dtype=dtypes)
